import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { EventEmitter } from '@angular/core';
import {
  ComponentFixture,
  TestBed,
  discardPeriodicTasks,
  fakeAsync,
  tick,
} from '@angular/core/testing';
import { Subscription } from 'rxjs';
import { InstantSearchComponent } from './instant-search.component';
import { InstantSearchQuery, InstantSearchResult } from './instant-search.model';

describe('InstantSearchComponent', () => {
  let component: InstantSearchComponent;
  let fixture: ComponentFixture<InstantSearchComponent>;

  const searchResults: InstantSearchResult<unknown>[] = [
    { title: 'test', description: 'test' },
    { title: 'caption', description: 'desc' },
  ];
  const searchBy: string = 'query';

  let searchQueryChanged: Mock<EventEmitter<any>>;
  let searchResultSelected: Mock<EventEmitter<any>>;

  beforeEach(async () => {
    searchQueryChanged = <any>mock(EventEmitter);
    searchResultSelected = <any>mock(EventEmitter);

    await TestBed.configureTestingModule({
      imports: [InstantSearchComponent],
    }).compileComponents();

    fixture = TestBed.createComponent(InstantSearchComponent);
    component = fixture.componentInstance;
    component.searchQueryChanged = useFromMock(searchQueryChanged);
    component.searchResultSelected = useFromMock(searchResultSelected);
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  describe('ngOnInit', () => {
    it('should handle value changes', () => {
      component.handleValueChanges = jest.fn();

      component.ngOnInit();

      expect(component.handleValueChanges).toHaveBeenCalled();
    });
  });

  describe('handleValueChanges', () => {
    beforeEach(() => {
      component.showResults = jest.fn();
    });

    it('should subscribe to value changes', () => {
      component.control.valueChanges.subscribe = jest.fn();

      component.handleValueChanges();

      expect(component.control.valueChanges.subscribe).toHaveBeenCalled();
    });

    it('should emit query', fakeAsync(() => {
      component.handleValueChanges();

      component.control.setValue(searchBy);

      tick(InstantSearchComponent.DEBOUNCE_TIME_IN_MILLIS);
      component.control.valueChanges.subscribe();

      expect(searchQueryChanged.emit).toHaveBeenCalledWith({ searchBy } as InstantSearchQuery);
    }));

    it('should not emit query', fakeAsync(() => {
      component.handleValueChanges();

      const searchBy: string = 'q';
      component.control.setValue(searchBy);

      tick(InstantSearchComponent.DEBOUNCE_TIME_IN_MILLIS);
      component.control.valueChanges.subscribe();

      expect(searchQueryChanged.emit).not.toHaveBeenCalled();
    }));

    describe('result are already visible', () => {
      beforeEach(() => {
        component.areResultsVisible = true;
      });

      it('should not show results', fakeAsync(() => {
        component.handleValueChanges();

        component.control.setValue(searchBy);
        tick(InstantSearchComponent.DEBOUNCE_TIME_IN_MILLIS);

        component.control.valueChanges.subscribe();
        expect(component.showResults).not.toHaveBeenCalled();

        discardPeriodicTasks();
      }));
    });

    describe('results are not visible', () => {
      beforeEach(() => {
        component.areResultsVisible = false;
      });

      it('should show results', fakeAsync(() => {
        component.handleValueChanges();

        component.control.setValue(searchBy);
        tick(InstantSearchComponent.DEBOUNCE_TIME_IN_MILLIS);

        component.control.valueChanges.subscribe();
        expect(component.showResults).toHaveBeenCalled();

        discardPeriodicTasks();
      }));

      it('should not show results if debounce time not reached', fakeAsync(() => {
        component.handleValueChanges();

        component.control.setValue(searchBy);
        tick(InstantSearchComponent.DEBOUNCE_TIME_IN_MILLIS - 1);

        component.control.valueChanges.subscribe();
        expect(component.showResults).not.toHaveBeenCalled();

        discardPeriodicTasks();
      }));

      it('should not show results if not enough characters entered', fakeAsync(() => {
        component.handleValueChanges();

        component.control.setValue('q');
        tick(InstantSearchComponent.DEBOUNCE_TIME_IN_MILLIS);

        component.control.valueChanges.subscribe();
        expect(component.showResults).not.toHaveBeenCalled();

        discardPeriodicTasks();
      }));
    });
  });

  describe('ngOnDestroy', () => {
    it('should subscribe to value changes', () => {
      component.formControlSubscription = new Subscription();
      component.formControlSubscription.unsubscribe = jest.fn();

      component.ngOnDestroy();

      expect(component.formControlSubscription.unsubscribe).toHaveBeenCalled();
    });
  });

  describe('set searchResults', () => {
    describe('on different results', () => {
      it('should call setSearchResults', () => {
        component.setSearchResults = jest.fn();

        component.searchResults = searchResults;

        expect(component.setSearchResults).toHaveBeenCalled();
      });
    });

    describe('on same results', () => {
      it('should not call setSearchResults', () => {
        component.setSearchResults = jest.fn();

        component.searchResults = [];

        expect(component.setSearchResults).not.toHaveBeenCalled();
      });
    });

    describe('on null or undefined', () => {
      it.each([null, undefined])(
        'should not call setSearchResults for %s',
        (searchResults: InstantSearchResult<unknown>[]) => {
          component.setSearchResults = jest.fn();

          component.searchResults = searchResults;

          expect(component.setSearchResults).not.toHaveBeenCalled();
        },
      );
    });
  });

  describe('setSearchResults', () => {
    it('should set results', () => {
      component.setSearchResults(searchResults);

      expect(component.results).toEqual(searchResults);
    });

    it('should call buildAriaLiveText with search results length', () => {
      component.buildAriaLiveText = jest.fn();

      component.setSearchResults(searchResults);

      expect(component.buildAriaLiveText).toHaveBeenCalledWith(searchResults.length);
    });
  });

  describe('setFocusOnResultItem', () => {
    beforeEach(() => {
      component.resultsRef.get = jest.fn().mockReturnValue({ setFocus: jest.fn() });
    });

    it('should call get for resultsRef with index', () => {
      component.setFocusOnResultItem(1);

      expect(component.resultsRef.get).toHaveBeenCalledWith(1);
    });

    it('should call setFocus', () => {
      component.setFocusOnResultItem(1);

      expect(component.resultsRef.get(1).setFocus).toHaveBeenCalled();
    });
  });

  describe('handleArrowNavigation', () => {
    const event: KeyboardEvent = new KeyboardEvent('arrow');

    beforeEach(() => {
      component.getResultIndexForKey = jest.fn();
      component.setFocusOnResultItem = jest.fn();
    });

    it('should call prevent default', () => {
      event.preventDefault = jest.fn();

      component.handleArrowNavigation(event);

      expect(event.preventDefault).toHaveBeenCalled();
    });

    it('should call getResultIndexForKey', () => {
      component.handleArrowNavigation(event);

      expect(component.getResultIndexForKey).toHaveBeenCalledWith(event.key);
    });

    it('should call setFocusOnResultItem', () => {
      component.getResultIndexForKey = jest.fn().mockReturnValue(0);

      component.handleArrowNavigation(event);

      expect(component.setFocusOnResultItem).toHaveBeenCalledWith(0);
    });
  });

  describe('handleEscape', () => {
    const event: KeyboardEvent = new KeyboardEvent('esc');

    it('should call prevent default', () => {
      event.preventDefault = jest.fn();

      component.handleEscape(event);

      expect(event.preventDefault).toHaveBeenCalled();
    });

    it('should call hideResults', () => {
      component.hideResults = jest.fn();

      component.handleEscape(event);

      expect(component.hideResults).toHaveBeenCalled();
    });
  });

  describe('getNextResultIndex', () => {
    it('should return 0 if index is undefined', () => {
      const result: number = component.getNextResultIndex(undefined, 2);

      expect(result).toBe(0);
    });

    it('should return 0 if current index is last', () => {
      const result: number = component.getNextResultIndex(1, 2);

      expect(result).toBe(0);
    });

    it('should return next search result index', () => {
      const result: number = component.getNextResultIndex(0, 2);

      expect(result).toBe(1);
    });
  });

  describe('getPreviousResultIndex', () => {
    it('should return last index if current index is undefined', () => {
      const result: number = component.getPreviousResultIndex(undefined, 2);

      expect(result).toBe(1);
    });

    it('should return last index if current index is first', () => {
      const result: number = component.getPreviousResultIndex(0, 2);

      expect(result).toBe(1);
    });

    it('should return previous search result index', () => {
      const result: number = component.getPreviousResultIndex(1, 2);

      expect(result).toBe(0);
    });
  });

  describe('getResultIndexForKey', () => {
    it('should call getNextResultIndex if ArrowDown', () => {
      component.getNextResultIndex = jest.fn();

      component.getResultIndexForKey('ArrowDown');

      expect(component.getNextResultIndex).toHaveBeenCalled();
    });

    it('should call getPreviousResultIndex if ArrowUp', () => {
      component.getPreviousResultIndex = jest.fn();

      component.getResultIndexForKey('ArrowUp');

      expect(component.getPreviousResultIndex).toHaveBeenCalled();
    });
  });

  describe('getLastItemIndex', () => {
    it('should return 0', () => {
      const result: number = component.getLastItemIndex(0);

      expect(result).toBe(0);
    });

    it('should return decrement of array length', () => {
      const result: number = component.getLastItemIndex(5);

      expect(result).toBe(4);
    });
  });

  describe('buildAriaLiveText', () => {
    beforeEach(() => {
      component.control.setValue('test');
    });

    it('should return text for one result', () => {
      const result: string = component.buildAriaLiveText(1);

      expect(result).toBe(
        'Ein Suchergebnis für Eingabe test. Nutze Pfeiltaste nach unten, um das zu erreichen.',
      );
    });

    it('should return text for many results', () => {
      const result: string = component.buildAriaLiveText(4);

      expect(result).toBe(
        '4 Suchergebnisse für Eingabe test. Nutze Pfeiltaste nach unten, um diese zu erreichen.',
      );
    });

    it('should return text for no results', () => {
      const result: string = component.buildAriaLiveText(0);

      expect(result).toBe('Keine Ergebnisse');
    });
  });

  describe('showResults', () => {
    it('should set isShowResults to true', () => {
      component.showResults();

      expect(component.areResultsVisible).toBe(true);
    });
  });

  describe('hideResults', () => {
    it('should set isShowResults to false', () => {
      component.hideResults();

      expect(component.areResultsVisible).toBe(false);
    });
  });

  describe('onKeydownHandler', () => {
    const keyboardEvent: KeyboardEvent = new KeyboardEvent('a');

    beforeEach(() => {
      component.isSearchResultsEmpty = jest.fn();
      component.isArrowNavigationKey = jest.fn();
      component.isEscapeKey = jest.fn();
      component.handleArrowNavigation = jest.fn();
      component.handleEscape = jest.fn();
    });

    it('should check for empty result', () => {
      component.onKeydownHandler(keyboardEvent);

      expect(component.isSearchResultsEmpty).toHaveBeenCalled();
    });

    describe('search result is empty', () => {
      beforeEach(() => {
        component.isSearchResultsEmpty = jest.fn().mockReturnValue(true);
      });

      it('should ignore key navigation', () => {
        component.onKeydownHandler(keyboardEvent);

        expect(component.isArrowNavigationKey).not.toHaveBeenCalled();
      });

      it('should ignore escape key handling', () => {
        component.onKeydownHandler(keyboardEvent);

        expect(component.isEscapeKey).not.toHaveBeenCalled();
      });
    });

    describe('search result is not empty', () => {
      beforeEach(() => {
        component.isSearchResultsEmpty = jest.fn().mockReturnValue(false);
      });

      it('should check if arrow navigation', () => {
        component.onKeydownHandler(keyboardEvent);

        expect(component.isArrowNavigationKey).toHaveBeenCalled();
      });

      it('should handle arrow navigation', () => {
        component.isArrowNavigationKey = jest.fn().mockReturnValue(true);

        component.onKeydownHandler(keyboardEvent);

        expect(component.handleArrowNavigation).toHaveBeenCalled();
      });

      it('should not handle arrow navigation', () => {
        component.isArrowNavigationKey = jest.fn().mockReturnValue(false);

        component.onKeydownHandler(keyboardEvent);

        expect(component.handleArrowNavigation).not.toHaveBeenCalled();
      });

      describe('is not arrow navigation', () => {
        beforeEach(() => {
          component.isArrowNavigationKey = jest.fn().mockReturnValue(false);
        });

        it('should check for escape key', () => {
          component.onKeydownHandler(keyboardEvent);

          expect(component.isEscapeKey).toHaveBeenCalled();
        });

        it('should handle escape key', () => {
          component.isEscapeKey = jest.fn().mockReturnValue(true);

          component.onKeydownHandler(keyboardEvent);

          expect(component.handleEscape).toHaveBeenCalled();
        });

        it('should not handle escape key', () => {
          component.isEscapeKey = jest.fn().mockReturnValue(false);

          component.onKeydownHandler(keyboardEvent);

          expect(component.handleEscape).not.toHaveBeenCalled();
        });
      });
    });
  });

  describe('isSearchResultsEmpty', () => {
    it('should return true', () => {
      component.results = [];

      const result: boolean = component.isSearchResultsEmpty();

      expect(result).toBe(true);
    });

    it('should return false', () => {
      component.results = searchResults;

      const result: boolean = component.isSearchResultsEmpty();

      expect(result).toBe(false);
    });
  });

  describe('isArrowNavigationKey', () => {
    it.each(['ArrowUp', 'ArrowDown'])('should return true for key %s', (key: string) => {
      const keyboardEvent: KeyboardEvent = { ...new KeyboardEvent('key'), key };

      const result: boolean = component.isArrowNavigationKey(keyboardEvent);

      expect(result).toBeTruthy();
    });

    it('should return false', () => {
      const result: boolean = component.isArrowNavigationKey(new KeyboardEvent('not arrow'));

      expect(result).toBe(false);
    });
  });

  describe('isEscapeKey', () => {
    it('should return true', () => {
      const escapeKeyEvent = { ...new KeyboardEvent('esc'), key: 'Escape' };

      const result: boolean = component.isEscapeKey(escapeKeyEvent);

      expect(result).toBe(true);
    });

    it('should return false', () => {
      const result: boolean = component.isEscapeKey(new KeyboardEvent('not escape'));

      expect(result).toBe(false);
    });
  });

  describe('isLastItemOrOutOfArray', () => {
    it.each([3, 5])('should return true for %s', (index: number) => {
      const result: boolean = component.isLastItemOrOutOfArray(index, 4);

      expect(result).toBe(true);
    });

    it('should return false', () => {
      const result: boolean = component.isLastItemOrOutOfArray(1, 3);

      expect(result).toBe(false);
    });
  });

  describe('isFirstItemOrOutOfArray', () => {
    it.each([0, -1])('should return true for %s', (index: number) => {
      const result: boolean = component.isFirstItemOrOutOfArray(index);

      expect(result).toBe(true);
    });

    it('should return false', () => {
      const result: boolean = component.isFirstItemOrOutOfArray(1);

      expect(result).toBe(false);
    });
  });

  describe('onItemClicked', () => {
    beforeEach(() => {
      component.hideResults = jest.fn();
    });

    it('should emit searchResultSelected', () => {
      component.onItemClicked(searchResults[0], 0);

      expect(searchResultSelected.emit).toHaveBeenCalledWith(searchResults[0]);
    });

    it('should hide results', () => {
      component.onItemClicked(searchResults[0], 0);

      expect(component.hideResults).toHaveBeenCalled();
    });
  });

  describe('onClickHandler', () => {
    const e: MouseEvent = { ...new MouseEvent('test') };

    beforeEach(() => {
      component.hideResults = jest.fn();
    });

    it('should call hideResults if instant search does not contain event target', () => {
      component.onClickHandler(e);

      expect(component.hideResults).toHaveBeenCalled();
    });

    it('should not call hideResults if instant search contains event target', () => {
      component.ref.nativeElement.contains = jest.fn().mockReturnValue(true);

      component.onClickHandler(e);

      expect(component.hideResults).not.toHaveBeenCalled();
    });
  });
});
