import { mock, useFromMock } from '@alfa-client/test-utils';
import { faker } from '@faker-js/faker';
import { Resource, ResourceFactory, ResourceUri, getUrl } from '@ngxp/rest';
import { DummyLinkRel, DummyListLinkRel } from 'libs/tech-shared/test/dummy';
import { createDummyListResource, createDummyResource } from 'libs/tech-shared/test/resource';
import { Observable } from 'rxjs';
import { singleCold, singleHot } from '../../../test/marbles';
import { CreateResourceData, LinkRelationName, SaveResourceData } from './resource.model';
import { ResourceRepository } from './resource.repository';
import { ListResource } from './resource.util';

describe('ResourceRepository', () => {
  let repository: ResourceRepository;

  let resourceFactory = mock(ResourceFactory);
  let resourceWrapper = { get: jest.fn(), post: jest.fn(), put: jest.fn(), delete: jest.fn() };

  beforeEach(() => {
    resourceFactory.from.mockReturnValue(resourceWrapper);
    resourceFactory.fromId.mockReturnValue(resourceWrapper);

    repository = new ResourceRepository(useFromMock(resourceFactory));
  });

  it('should be created', () => {
    expect(repository).toBeTruthy();
  });

  describe('get list resource', () => {
    const listLinkRel: string = DummyListLinkRel.LIST;
    const baseResource: Resource = createDummyResource([listLinkRel]);

    const loadedListResource: ListResource = createDummyListResource();
    const uriWithoutParameter: string = faker.internet.url();

    beforeEach(() => {
      resourceWrapper.get.mockReturnValue(singleCold(loadedListResource));

      repository.getUrlWithoutParameter = jest.fn().mockReturnValue(uriWithoutParameter);
    });

    it('should call get url without parameter', () => {
      repository.getListResource(baseResource, listLinkRel);

      expect(repository.getUrlWithoutParameter).toHaveBeenCalledWith(
        getUrl(baseResource, listLinkRel),
      );
    });

    it('should call resourceFactory with uri', () => {
      repository.getListResource(baseResource, listLinkRel);

      expect(resourceFactory.fromId).toHaveBeenCalledWith(uriWithoutParameter);
    });

    it('should call resourceWrapper', () => {
      repository.getListResource(baseResource, listLinkRel);

      expect(resourceWrapper.get).toHaveBeenCalled();
    });

    it('should return result', () => {
      const result: Observable<Resource> = repository.getListResource(baseResource, listLinkRel);

      expect(result).not.toBeNull();
      expect(result).toBeObservable(singleHot(loadedListResource));
    });
  });

  describe('get url without parameter', () => {
    const baseUri: string = faker.internet.url();

    it('should remove parameter', () => {
      const uri: ResourceUri = baseUri + '{?page,size,sort}';

      const uriWithoutParameter: string = repository.getUrlWithoutParameter(uri);

      expect(uriWithoutParameter).toEqual(baseUri);
    });

    it('should return original uri if no parameter exists', () => {
      const uriWithoutParameter: string = repository.getUrlWithoutParameter(baseUri);

      expect(uriWithoutParameter).toEqual(baseUri);
    });
  });

  describe('get resource', () => {
    const uri: ResourceUri = faker.internet.url();
    const loadedResource: Resource = createDummyResource();

    beforeEach(() => {
      resourceWrapper.get.mockReturnValue(singleCold(loadedResource));
    });

    it('should call resourceFactory with resource', () => {
      repository.getResource(uri);

      expect(resourceFactory.fromId).toHaveBeenCalledWith(uri);
    });

    it('should call resourceWrapper with link', () => {
      repository.getResource(uri);

      expect(resourceWrapper.get).toHaveBeenCalled();
    });

    it('should return result', () => {
      const result: Observable<Resource> = repository.getResource(uri);

      expect(result).not.toBeNull();
      expect(result).toBeObservable(singleHot(loadedResource));
    });
  });

  describe('create resource', () => {
    const toCreate: any = {};
    const linkRel: string = DummyLinkRel.DUMMY;
    const resource: Resource = createDummyResource([linkRel]);
    const createResourceData: CreateResourceData<Resource> = { toCreate, linkRel, resource };

    const createdResource: Resource = createDummyResource();

    beforeEach(() => {
      resourceWrapper.post.mockReturnValue(singleCold(createdResource));
    });

    it('should call resourceFactory with resource', () => {
      repository.createResource(createResourceData);

      expect(resourceFactory.from).toHaveBeenCalledWith(resource);
    });

    it('should call resourceWrapper with linkel and object to create', () => {
      repository.createResource(createResourceData);

      expect(resourceWrapper.post).toHaveBeenCalledWith(linkRel, toCreate);
    });

    it('should return value', () => {
      const result: Observable<Resource> = repository.createResource(createResourceData);

      expect(result).toBeObservable(singleHot(createdResource));
    });
  });

  describe('save', () => {
    const toSave: unknown = {};
    const linkRel: string = DummyLinkRel.DUMMY;
    const resource: Resource = createDummyResource([linkRel]);
    const saveResourceData: SaveResourceData<Resource> = { toSave, linkRel, resource };

    const savedResource: Resource = createDummyResource();

    beforeEach(() => {
      resourceWrapper.put.mockReturnValue(singleCold(savedResource));
    });

    it('should call resourceFactory with resource', () => {
      repository.save(saveResourceData);

      expect(resourceFactory.from).toHaveBeenCalledWith(resource);
    });

    it('should call resourceWrapper with linkel and object to save', () => {
      repository.save(saveResourceData);

      expect(resourceWrapper.put).toHaveBeenCalledWith(linkRel, toSave);
    });

    it('should return value', () => {
      const result: Observable<Resource> = repository.save(saveResourceData);

      expect(result).toBeObservable(singleHot(savedResource));
    });
  });

  describe('delete', () => {
    const deleteLinkRel: LinkRelationName = faker.random.word();
    const resourceToDelete: Resource = createDummyResource([deleteLinkRel]);

    const deletedResource: Resource = createDummyResource();

    beforeEach(() => {
      resourceWrapper.delete.mockReturnValue(singleCold(deletedResource));
    });

    it('should call resourceFactory with resource', () => {
      repository.delete(resourceToDelete, deleteLinkRel);

      expect(resourceFactory.from).toHaveBeenCalledWith(resourceToDelete);
    });

    it('should call resourceWrapper with linkel', () => {
      repository.delete(resourceToDelete, deleteLinkRel);

      expect(resourceWrapper.delete).toHaveBeenCalledWith(deleteLinkRel);
    });

    it('should return value', () => {
      const result: Observable<Resource> = repository.delete(resourceToDelete, deleteLinkRel);

      expect(result).toBeObservable(singleHot(deletedResource));
    });
  });
});
