/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.common.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Calendar;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;

import de.ozgcloud.alfa.loeschanforderung.DeleteLoeschAnforderung;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;

class CommandServiceTest {

	@Spy
	@InjectMocks
	private CommandService service;
	@Mock
	private CommandRemoteService remoteService;

	@Captor
	private ArgumentCaptor<CreateCommand> createCommandCaptor;

	@Nested
	class TestCreateCommand {

		private final CreateCommand command = CommandTestFactory.createCreateCommand();

		@Nested
		class TestWithCreateCommandOnly {

			@Test
			void shouldCallRemoteService() {
				service.createCommandDeprecated(command);

				verify(remoteService).createCommand(any(CreateCommand.class));
			}

			@Test
			void shouldHaveSetRelationVersion() {
				service.createCommandDeprecated(command);

				verify(remoteService).createCommand(createCommandCaptor.capture());
				assertThat(createCommandCaptor.getValue().getRelationVersion()).isEqualTo(CommandService.NO_RELATION_VERSION);
			}
		}

		@Nested
		class TestWithRelationVersion {

			@Test
			void shouldCallRemoteService() {
				service.createCommand(command, CommandTestFactory.RELATION_VERSION);

				verify(remoteService).createCommand(any(CreateCommand.class));
			}

			@Test
			void shouldHaveSetRelationVersion() {
				service.createCommand(command, CommandTestFactory.RELATION_VERSION);

				verify(remoteService).createCommand(createCommandCaptor.capture());
				assertThat(createCommandCaptor.getValue().getRelationVersion()).isEqualTo(CommandTestFactory.RELATION_VERSION);
			}
		}
	}

	@Nested
	class TestCreateCommandWithItemName {

		private final CreateCommand command = CommandTestFactory.createCreateCommand();

		@Nested
		class WithoutValidation {

			@Test
			void shouldCallRemoteService() {
				service.createCommand(command, CommandTestFactory.RELATION_VERSION);

				verify(remoteService).createCommand(any(CreateCommand.class));
			}

			@Test
			void shouldHaveSetRelationVersion() {
				service.createCommand(command, CommandTestFactory.RELATION_VERSION);

				verify(remoteService).createCommand(createCommandCaptor.capture());
				assertThat(createCommandCaptor.getValue().getRelationVersion()).isEqualTo(CommandTestFactory.RELATION_VERSION);
			}
		}
	}

	@Nested
	class TestGetById {

		@BeforeEach
		void init() {
			when(remoteService.getCommand(CommandTestFactory.ID)).thenReturn(CommandTestFactory.create());
		}

		@Test
		void shouldCallRemoteService() {
			service.getById(CommandTestFactory.ID);

			verify(remoteService).getCommand(CommandTestFactory.ID);
			verify(service, never()).handleDeleteLoeschAnforderung(any());
		}

		@DisplayName("As deleteLoeschAnforderung")
		@Nested
		class TestAsDeleteLoeschAnforderungCommand {

			private final Command command = CommandTestFactory.createBuilder().order(CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN.name())
					.build();

			@BeforeEach
			void init() {
				when(remoteService.getCommand(CommandTestFactory.ID)).thenReturn(command);
			}

			@Test
			void shouldMarkAsDone() {
				doNothing().when(service).handleDeleteLoeschAnforderung(any());

				service.getById(CommandTestFactory.ID);

				verify(service).handleDeleteLoeschAnforderung(command);
			}
		}
	}

	@DisplayName("Handle deleteLoeschAnforderung")
	@Nested
	class TestHandleDeleteLoeschAnforderung {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallIsDeletedAttachedItemCommandDone() {
			doReturn(false).when(service).isDeleteAttachedItemCommandFinished(any());

			service.handleDeleteLoeschAnforderung(command);

			verify(service).isDeleteAttachedItemCommandFinished(command);
		}

		@Test
		void shouldCallIsChangeStatusCommandDone() {
			doReturn(true).when(service).isDeleteAttachedItemCommandFinished(any());
			doReturn(false).when(service).isChangeStatusCommandFinished(any());

			service.handleDeleteLoeschAnforderung(command);

			verify(service).isChangeStatusCommandFinished(command);
		}

		@Test
		void shouldCallRemoteServiceIfCommandsFinisihed() {
			doReturn(true).when(service).isDeleteAttachedItemCommandFinished(any());
			doReturn(true).when(service).isChangeStatusCommandFinished(any());

			service.handleDeleteLoeschAnforderung(command);

			verify(remoteService).finishCommand(command.getId());
		}

		@DisplayName("Is deleteAttachedItemCommand finished")
		@Nested
		class TestIsDeletedAttachedItemCommandFinished {

			private final String deleteAttachedItemCommandId = UUID.randomUUID().toString();
			private final Command command = CommandTestFactory.createBuilder()
					.body(Map.of(DeleteLoeschAnforderung.DELETE_ATTACHED_ITEM_COMMAND_ID_FIELD, deleteAttachedItemCommandId)).build();

			@Test
			void shouldGetCommandById() {
				when(remoteService.getCommand(any())).thenReturn(CommandTestFactory.create());

				service.isDeleteAttachedItemCommandFinished(command);

				verify(remoteService).getCommand(deleteAttachedItemCommandId);
			}

			@Test
			void shouldReturnTrueIfCommandIsFinished() {
				when(remoteService.getCommand(any())).thenReturn(CommandTestFactory.createBuilder().status(CommandStatus.FINISHED).build());

				var isFinished = service.isDeleteAttachedItemCommandFinished(command);

				assertThat(isFinished).isTrue();
			}
		}

		@DisplayName("Is changeStatusCommand finished")
		@Nested
		class TestIsChangeStatusCommandFinished {

			private final String changeStatusCommandId = UUID.randomUUID().toString();
			private final Command command = CommandTestFactory.createBuilder()
					.body(Map.of(DeleteLoeschAnforderung.CHANGE_STATUS_COMMAND_ID_FIELD, changeStatusCommandId)).build();

			@Test
			void shouldGetCommandById() {
				when(remoteService.getCommand(any())).thenReturn(CommandTestFactory.create());

				service.isChangeStatusCommandFinished(command);

				verify(remoteService).getCommand(changeStatusCommandId);
			}

			@Test
			void shouldReturnTrueIfCommandIsFinished() {
				when(remoteService.getCommand(any())).thenReturn(CommandTestFactory.createBuilder().status(CommandStatus.FINISHED).build());

				var isFinished = service.isChangeStatusCommandFinished(command);

				assertThat(isFinished).isTrue();
			}
		}
	}

	@Nested
	class TestRevoke {

		@Test
		void shouldCallRemoteService() {
			service.revoke(CommandTestFactory.RELATION_ID);

			verify(remoteService).revokeCommand(CommandTestFactory.RELATION_ID);
		}
	}

	@Nested
	class TestExistsPendingCommands {

		@Test
		void shouldCallRemoteService() {
			service.existsPendingCommands(VorgangHeaderTestFactory.ID);

			verify(remoteService).existsPendingCommands(VorgangHeaderTestFactory.ID);
		}
	}

	@Nested
	class TestGetPendingCommands {

		@Test
		void shouldCallRemoteService() {
			service.getPendingCommands(VorgangHeaderTestFactory.ID);

			verify(remoteService).getPendingCommands(VorgangHeaderTestFactory.ID);
		}
	}

	@DisplayName("Find finished commands")
	@Nested
	class TestFindFinishedCommands {

		@Test
		void shouldCallRemoteService() {
			service.findFinishedCommands(VorgangHeaderTestFactory.ID);

			verify(remoteService).findCommands(VorgangHeaderTestFactory.ID, Optional.of(CommandStatus.FINISHED), Optional.empty());
		}
	}

	@Nested
	class TestWaitUntilDone {

		private final Command pendingCommand = CommandTestFactory.createBuilder().status(CommandStatus.PENDING).build();
		private final Command finishedCommand = pendingCommand.toBuilder().status(CommandStatus.FINISHED).build();

		@Nested
		class OnFinishedCommand {

			@Test
			void shouldReturnFinishedCommand() {
				var resultCommand = service.waitUntilDone(finishedCommand);

				assertThat(resultCommand).isEqualTo(finishedCommand);
			}

			@Test
			void shouldNotReloadCommand() {
				service.waitUntilDone(finishedCommand);

				verify(service, never()).reloadCommand(any());
			}
		}

		@Nested
		class OnPendingCommand {

			@BeforeEach
			void setUp() {
				doReturn(finishedCommand).when(service).reloadCommand(pendingCommand.getId());
			}

			@Test
			void shouldReloadCommand() {
				service.waitUntilDone(pendingCommand);

				verify(service).reloadCommand(pendingCommand.getId());
			}

			@Test
			void shouldReturnDoneCommand() {
				var resultCommand = service.waitUntilDone(pendingCommand);

				assertThat(resultCommand).isEqualTo(finishedCommand);
			}
		}

		@Nested
		class OnTimeoutExceeded {

			@Mock
			private Calendar calendar;

			@Test
			void shouldReturnPendingCommand() {
				try (MockedStatic<Calendar> calendarMockedStatic = mockStatic(Calendar.class)) {
					calendarMockedStatic.when(Calendar::getInstance).thenReturn(calendar);
					when(calendar.getTimeInMillis()).thenReturn(0L, 15000L);

					var resultCommand = service.waitUntilDone(pendingCommand);

					assertThat(resultCommand).isEqualTo(pendingCommand);
				}
			}

			@Test
			void shouldReloadPendingCommand() {
				doReturn(pendingCommand).when(service).reloadCommand(pendingCommand.getId());

				try (MockedStatic<Calendar> calendarMockedStatic = mockStatic(Calendar.class)) {
					calendarMockedStatic.when(Calendar::getInstance).thenReturn(calendar);
					when(calendar.getTimeInMillis()).thenReturn(0L, 0L, 15000L);

					service.waitUntilDone(pendingCommand);

					verify(service).reloadCommand(pendingCommand.getId());
				}
			}

			@Test
			void shouldReturnPendingCommandAfterReload() {
				doReturn(pendingCommand).when(service).reloadCommand(pendingCommand.getId());

				try (MockedStatic<Calendar> calendarMockedStatic = mockStatic(Calendar.class)) {
					calendarMockedStatic.when(Calendar::getInstance).thenReturn(calendar);
					when(calendar.getTimeInMillis()).thenReturn(0L, 0L, 15000L);

					var resultCommand = service.waitUntilDone(pendingCommand);

					assertThat(resultCommand).isEqualTo(pendingCommand);
				}
			}
		}
	}
}